import copy
import math
import numbers
from numbers import Integral, Real

import numpy as np
from scipy.sparse import issparse
from sklearn.utils import check_random_state
from sklearn.utils._param_validation import Interval, RealNotInt, StrOptions

from .._lib.sklearn.tree import DecisionTreeClassifier, _criterion
from .._lib.sklearn.tree import _tree as _sklearn_tree
from .._lib.sklearn.tree._criterion import BaseCriterion
from .._lib.sklearn.tree._tree import BestFirstTreeBuilder, DepthFirstTreeBuilder
from . import _oblique_splitter
from ._neighbors import SimMatrixMixin
from ._oblique_splitter import ObliqueSplitter
from ._oblique_tree import ObliqueTree

DTYPE = _sklearn_tree.DTYPE
DOUBLE = _sklearn_tree.DOUBLE

CRITERIA_CLF = {
    "gini": _criterion.Gini,
    "log_loss": _criterion.Entropy,
    "entropy": _criterion.Entropy,
}
CRITERIA_REG = {
    "squared_error": _criterion.MSE,
    "friedman_mse": _criterion.FriedmanMSE,
    "absolute_error": _criterion.MAE,
    "poisson": _criterion.Poisson,
}

DENSE_SPLITTERS = {
    "best": _oblique_splitter.MultiViewSplitter,
}


class MultiViewDecisionTreeClassifier(SimMatrixMixin, DecisionTreeClassifier):
    """A multi-view axis-aligned decision tree classifier.

    This is an experimental feature that applies an oblique decision tree to
    multiple feature-sets concatenated across columns in ``X``.

    Parameters
    ----------
    criterion : {"gini", "entropy"}, default="gini"
        The function to measure the quality of a split. Supported criteria are
        "gini" for the Gini impurity and "entropy" for the information gain.

    splitter : {"best"}, default="best"
        The strategy used to choose the split at each node.

    max_depth : int, default=None
        The maximum depth of the tree. If None, then nodes are expanded until
        all leaves are pure or until all leaves contain less than
        min_samples_split samples.

    min_samples_split : int or float, default=2
        The minimum number of samples required to split an internal node:

        - If int, then consider `min_samples_split` as the minimum number.
        - If float, then `min_samples_split` is a fraction and
          `ceil(min_samples_split * n_samples)` are the minimum
          number of samples for each split.

    min_samples_leaf : int or float, default=1
        The minimum number of samples required to be at a leaf node.
        A split point at any depth will only be considered if it leaves at
        least ``min_samples_leaf`` training samples in each of the left and
        right branches.  This may have the effect of smoothing the model,
        especially in regression.

        - If int, then consider `min_samples_leaf` as the minimum number.
        - If float, then `min_samples_leaf` is a fraction and
          `ceil(min_samples_leaf * n_samples)` are the minimum
          number of samples for each node.

    min_weight_fraction_leaf : float, default=0.0
        The minimum weighted fraction of the sum total of weights (of all
        the input samples) required to be at a leaf node. Samples have
        equal weight when sample_weight is not provided.

    max_features : array-like, int, float or {"auto", "sqrt", "log2"}, default=None
        The number of features to consider when looking for the best split:

            - If int, then consider `max_features` features at each split.
            - If float, then `max_features` is a fraction and
              `int(max_features * n_features)` features are considered at each
              split.
            - If "auto", then `max_features=sqrt(n_features)`.
            - If "sqrt", then `max_features=sqrt(n_features)`.
            - If "log2", then `max_features=log2(n_features)`.
            - If None, then `max_features=n_features`.

        If array-like, then `max_features` is the number of features to consider
        for each feature set following the same logic as above, where
        ``n_features`` is the number of features in the respective feature set.

        Note: the search for a split does not stop until at least one
        valid partition of the node samples is found, even if it requires to
        effectively inspect more than ``max_features`` features.

        Note: Compared to axis-aligned Random Forests, one can set
        max_features to a number greater then ``n_features``.

    random_state : int, RandomState instance or None, default=None
        Controls the randomness of the estimator. The features are always
        randomly permuted at each split, even if ``splitter`` is set to
        ``"best"``. When ``max_features < n_features``, the algorithm will
        select ``max_features`` at random at each split before finding the best
        split among them. But the best found split may vary across different
        runs, even if ``max_features=n_features``. That is the case, if the
        improvement of the criterion is identical for several splits and one
        split has to be selected at random. To obtain a deterministic behaviour
        during fitting, ``random_state`` has to be fixed to an integer.
        See :term:`Glossary <random_state>` for details.

    max_leaf_nodes : int, default=None
        Grow a tree with ``max_leaf_nodes`` in best-first fashion.
        Best nodes are defined as relative reduction in impurity.
        If None then unlimited number of leaf nodes.

    min_impurity_decrease : float, default=0.0
        A node will be split if this split induces a decrease of the impurity
        greater than or equal to this value.

        The weighted impurity decrease equation is the following::

            N_t / N * (impurity - N_t_R / N_t * right_impurity
                                - N_t_L / N_t * left_impurity)

        where ``N`` is the total number of samples, ``N_t`` is the number of
        samples at the current node, ``N_t_L`` is the number of samples in the
        left child, and ``N_t_R`` is the number of samples in the right child.

        ``N``, ``N_t``, ``N_t_R`` and ``N_t_L`` all refer to the weighted sum,
        if ``sample_weight`` is passed.

    class_weight : dict, list of dict or "balanced", default=None
        Weights associated with classes in the form ``{class_label: weight}``.
        If None, all classes are supposed to have weight one. For
        multi-output problems, a list of dicts can be provided in the same
        order as the columns of y.

        Note that for multioutput (including multilabel) weights should be
        defined for each class of every column in its own dict. For example,
        for four-class multilabel classification weights should be
        [{0: 1, 1: 1}, {0: 1, 1: 5}, {0: 1, 1: 1}, {0: 1, 1: 1}] instead of
        [{1:1}, {2:5}, {3:1}, {4:1}].

        The "balanced" mode uses the values of y to automatically adjust
        weights inversely proportional to class frequencies in the input data
        as ``n_samples / (n_classes * np.bincount(y))``

        For multi-output, the weights of each column of y will be multiplied.

        Note that these weights will be multiplied with sample_weight (passed
        through the fit method) if sample_weight is specified.

    feature_combinations : float, default=None
        Not used.

    ccp_alpha : non-negative float, default=0.0
        Not used.

    store_leaf_values : bool, default=False
        Whether to store the leaf values.

    monotonic_cst : array-like of int of shape (n_features), default=None
        Indicates the monotonicity constraint to enforce on each feature.
          - 1: monotonic increase
          - 0: no constraint
          - -1: monotonic decrease

        Not used.

    feature_set_ends : array-like of int of shape (n_feature_sets,), default=None
        The indices of the end of each feature set. For example, if the first
        feature set is the first 10 features, and the second feature set is the
        next 20 features, then ``feature_set_ends = [10, 30]``. If ``None``,
        then this will assume that there is only one feature set.

    apply_max_features_per_feature_set : bool, default=False
        Whether to apply sampling per feature set, where ``max_features`` is applied
        to each feature-set. If ``False``, then sampling
        is applied over the entire feature space.

    Attributes
    ----------
    classes_ : ndarray of shape (n_classes,) or list of ndarray
        The classes labels (single output problem),
        or a list of arrays of class labels (multi-output problem).

    feature_importances_ : ndarray of shape (n_features,)
        The impurity-based feature importances.
        The higher, the more important the feature.
        The importance of a feature is computed as the (normalized)
        total reduction of the criterion brought by that feature.  It is also
        known as the Gini importance [4]_.

        Warning: impurity-based feature importances can be misleading for
        high cardinality features (many unique values). See
        :func:`sklearn.inspection.permutation_importance` as an alternative.

    max_features_ : int
        The inferred value of max_features.

    n_classes_ : int or list of int
        The number of classes (for single output problems),
        or a list containing the number of classes for each
        output (for multi-output problems).

    n_features_in_ : int
        Number of features seen during :term:`fit`.

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of features seen during :term:`fit`. Defined only when `X`
        has feature names that are all strings.

    n_outputs_ : int
        The number of outputs when ``fit`` is performed.

    tree_ : Tree instance
        The underlying Tree object. Please refer to
        ``help(sklearn.tree._tree.Tree)`` for
        attributes of Tree object.

    feature_combinations_ : float
        The number of feature combinations on average taken to fit the tree.

    feature_set_ends_ : array-like of int of shape (n_feature_sets,)
        The indices of the end of each feature set.

    n_feature_sets_ : int
        The number of feature sets.

    max_features_per_set_ : array-like of int of shape (n_feature_sets,)
        The number of features to sample per feature set. If ``None``, then
        ``max_features`` is applied to the entire feature space.

    See Also
    --------
    sklearn.tree.DecisionTreeClassifier : An axis-aligned decision tree classifier.
    """

    tree_type = "oblique"

    _parameter_constraints = {
        **DecisionTreeClassifier._parameter_constraints,
        "feature_combinations": [
            Interval(Real, 1.0, None, closed="left"),
            None,
        ],
        "feature_set_ends": ["array-like", None],
        "apply_max_features_per_feature_set": ["boolean"],
    }
    _parameter_constraints.pop("max_features")
    _parameter_constraints["max_features"] = [
        Interval(Integral, 1, None, closed="left"),
        Interval(RealNotInt, 0.0, 1.0, closed="right"),
        StrOptions({"sqrt", "log2"}),
        "array-like",
        None,
    ]

    def __init__(
        self,
        *,
        criterion="gini",
        splitter="best",
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_weight_fraction_leaf=0.0,
        max_features=None,
        random_state=None,
        max_leaf_nodes=None,
        min_impurity_decrease=0.0,
        class_weight=None,
        feature_combinations=None,
        ccp_alpha=0.0,
        store_leaf_values=False,
        monotonic_cst=None,
        feature_set_ends=None,
        apply_max_features_per_feature_set=False,
    ):
        super().__init__(
            criterion=criterion,
            splitter=splitter,
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_weight_fraction_leaf=min_weight_fraction_leaf,
            max_features=max_features,
            max_leaf_nodes=max_leaf_nodes,
            class_weight=class_weight,
            random_state=random_state,
            min_impurity_decrease=min_impurity_decrease,
            ccp_alpha=ccp_alpha,
            store_leaf_values=store_leaf_values,
            monotonic_cst=monotonic_cst,
        )

        self.feature_combinations = feature_combinations
        self.feature_set_ends = feature_set_ends
        self.apply_max_features_per_feature_set = apply_max_features_per_feature_set
        self._max_features_arr = None

    def _build_tree(
        self,
        X,
        y,
        sample_weight,
        missing_values_in_feature_mask,
        min_samples_leaf,
        min_weight_leaf,
        max_leaf_nodes,
        min_samples_split,
        max_depth,
        random_state,
    ):
        """Build the actual tree.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The training input samples. Internally, it will be converted to
            ``dtype=np.float32`` and if a sparse matrix is provided
            to a sparse ``csc_matrix``.

        y : array-like of shape (n_samples,) or (n_samples, n_outputs)
            The target values (class labels) as integers or strings.

        sample_weight : array-like of shape (n_samples,), default=None
            Sample weights. If None, then samples are equally weighted. Splits
            that would create child nodes with net zero or negative weight are
            ignored while searching for a split in each node. Splits are also
            ignored if they would result in any single class carrying a
            negative weight in either child node.

        min_samples_leaf : int or float
            The minimum number of samples required to be at a leaf node.

        min_weight_leaf : float, default=0.0
           The minimum weighted fraction of the sum total of weights.

        max_leaf_nodes : int, default=None
            Grow a tree with ``max_leaf_nodes`` in best-first fashion.

        min_samples_split : int or float, default=2
            The minimum number of samples required to split an internal node.

        max_depth : int, default=None
            The maximum depth of the tree. If None, then nodes are expanded until
            all leaves are pure or until all leaves contain less than
            min_samples_split samples.

        random_state : int, RandomState instance or None, default=None
            Controls the randomness of the estimator.
        """
        monotonic_cst = None
        self.monotonic_cst_ = monotonic_cst
        _, n_features = X.shape

        self.feature_combinations_ = 1

        # Build tree
        criterion = self.criterion
        if not isinstance(criterion, BaseCriterion):
            criterion = CRITERIA_CLF[self.criterion](self.n_outputs_, self.n_classes_)
        else:
            # Make a deepcopy in case the criterion has mutable attributes that
            # might be shared and modified concurrently during parallel fitting
            criterion = copy.deepcopy(criterion)

        if self.feature_set_ends is None:
            self.feature_set_ends_ = np.asarray([n_features], dtype=np.intp)
        else:
            self.feature_set_ends_ = np.atleast_1d(self.feature_set_ends).astype(np.intp)
        self.n_feature_sets_ = len(self.feature_set_ends_)
        if self.feature_set_ends_[-1] != n_features:
            raise ValueError(
                f"The last feature set end must be equal to the number of features, "
                f"{n_features}, but got {self.feature_set_ends_[-1]}."
            )

        splitter = self.splitter
        if issparse(X):
            raise ValueError(
                "Sparse input is not supported for oblique trees. "
                "Please convert your data to a dense array."
            )
        else:
            SPLITTERS = DENSE_SPLITTERS

        if isinstance(self._max_features_arr, (Integral, Real, str, type(None))):
            max_features_arr_ = [self._max_features_arr] * self.n_feature_sets_
            stratify_mtry_per_view = self.apply_max_features_per_feature_set
        else:
            if not isinstance(self._max_features_arr, (list, np.ndarray)):
                raise ValueError(
                    f"max_features must be an array-like, int, float, str, or None; "
                    f"got {type(self._max_features_arr)}"
                )
            if len(self._max_features_arr) != self.n_feature_sets_:
                raise ValueError(
                    f"max_features must be an array-like of length {self.n_feature_sets_}; "
                    f"got {len(self.max_features)}"
                )
            max_features_arr_ = self._max_features_arr
            stratify_mtry_per_view = True

        self.n_features_in_set_ = []
        if stratify_mtry_per_view:
            # XXX: experimental
            # we can replace max_features_ here based on whether or not uniform logic over
            # feature sets
            max_features_per_set = []
            n_features_in_prev = 0
            for idx in range(self.n_feature_sets_):
                max_features = max_features_arr_[idx]

                n_features_in_ = self.feature_set_ends_[idx] - n_features_in_prev
                n_features_in_prev += n_features_in_
                self.n_features_in_set_.append(n_features_in_)
                if isinstance(max_features, str):
                    if max_features == "sqrt":
                        max_features = max(1, math.ceil(np.sqrt(n_features_in_)))
                    elif max_features == "log2":
                        max_features = max(1, math.ceil(np.log2(n_features_in_)))
                elif max_features is None:
                    max_features = n_features_in_
                elif isinstance(max_features, numbers.Integral):
                    max_features = max_features
                else:  # float
                    if max_features > 0.0:
                        max_features = max(1, math.ceil(max_features * n_features_in_))
                    else:
                        max_features = 0

                if max_features > n_features_in_:
                    raise ValueError(
                        f"max_features must be less than or equal to "
                        f"the number of features in feature set {idx}: {n_features_in_}, but "
                        f"max_features = {max_features} when applying sampling"
                        f"per feature set."
                    )

                max_features_per_set.append(max_features)
            self.max_features_ = np.sum(max_features_per_set)
            if self.max_features_ > n_features:
                raise ValueError(
                    "max_features is greater than the number of features: "
                    f"{max_features} > {n_features}."
                    "This should not be possible. Please submit a bug report."
                )
            self.max_features_per_set_ = np.asarray(max_features_per_set, dtype=np.intp)
            # the total number of features to sample per split
            self.max_features_ = np.sum(self.max_features_per_set_)
        else:
            self.max_features_per_set_ = None
            self.max_features = self._max_features_arr
            if isinstance(self.max_features, str):
                if self.max_features == "sqrt":
                    max_features = max(1, int(np.sqrt(self.n_features_in_)))
                elif self.max_features == "log2":
                    max_features = max(1, int(np.log2(self.n_features_in_)))
            elif self.max_features is None:
                max_features = self.n_features_in_
            elif isinstance(self.max_features, numbers.Integral):
                max_features = self.max_features
            else:  # float
                if self.max_features > 0.0:
                    max_features = max(1, int(self.max_features * self.n_features_in_))
                else:
                    max_features = 0

            self.max_features_ = max_features

        if not isinstance(self.splitter, ObliqueSplitter):
            splitter = SPLITTERS[self.splitter](
                criterion,
                self.max_features_,
                min_samples_leaf,
                min_weight_leaf,
                random_state,
                monotonic_cst,
                self.feature_combinations_,
                self.feature_set_ends_,
                self.n_feature_sets_,
                self.max_features_per_set_,
            )

        self.tree_ = ObliqueTree(self.n_features_in_, self.n_classes_, self.n_outputs_)

        # Use BestFirst if max_leaf_nodes given; use DepthFirst otherwise
        if max_leaf_nodes < 0:
            builder = DepthFirstTreeBuilder(
                splitter,
                min_samples_split,
                min_samples_leaf,
                min_weight_leaf,
                max_depth,
                self.min_impurity_decrease,
            )
        else:
            builder = BestFirstTreeBuilder(
                splitter,
                min_samples_split,
                min_samples_leaf,
                min_weight_leaf,
                max_depth,
                max_leaf_nodes,
                self.min_impurity_decrease,
            )

        builder.build(self.tree_, X, y, sample_weight, None)

        if self.n_outputs_ == 1:
            self.n_classes_ = self.n_classes_[0]
            self.classes_ = self.classes_[0]

    def _update_tree(self, X, y, sample_weight):
        # Update tree
        max_leaf_nodes = -1 if self.max_leaf_nodes is None else self.max_leaf_nodes
        min_samples_split = self.min_samples_split_
        min_samples_leaf = self.min_samples_leaf_
        min_weight_leaf = self.min_weight_leaf_
        # set decision-tree model parameters
        max_depth = np.iinfo(np.int32).max if self.max_depth is None else self.max_depth

        monotonic_cst = self.monotonic_cst_

        # Build tree
        # Note: this reconstructs the builder with the same state it had during the
        # initial fit. This is necessary because the builder is not saved as part
        # of the class, and thus the state may be lost if pickled/unpickled.
        criterion = self.criterion
        if not isinstance(criterion, BaseCriterion):
            criterion = CRITERIA_CLF[self.criterion](self.n_outputs_, self._n_classes_)
        else:
            # Make a deepcopy in case the criterion has mutable attributes that
            # might be shared and modified concurrently during parallel fitting
            criterion = copy.deepcopy(criterion)

        random_state = check_random_state(self.random_state)

        splitter = self.splitter
        if issparse(X):
            raise ValueError(
                "Sparse input is not supported for oblique trees. "
                "Please convert your data to a dense array."
            )
        else:
            SPLITTERS = DENSE_SPLITTERS
        if not isinstance(self.splitter, ObliqueSplitter):
            splitter = SPLITTERS[self.splitter](
                criterion,
                self.max_features_,
                min_samples_leaf,
                min_weight_leaf,
                random_state,
                monotonic_cst,
                self.feature_combinations_,
                self.feature_set_ends_,
                self.n_feature_sets_,
                self.max_features_per_set_,
            )

        # Use BestFirst if max_leaf_nodes given; use DepthFirst otherwise
        if max_leaf_nodes < 0:
            builder = DepthFirstTreeBuilder(
                splitter,
                min_samples_split,
                min_samples_leaf,
                min_weight_leaf,
                max_depth,
                self.min_impurity_decrease,
                self.store_leaf_values,
            )
        else:
            builder = BestFirstTreeBuilder(
                splitter,
                min_samples_split,
                min_samples_leaf,
                min_weight_leaf,
                max_depth,
                max_leaf_nodes,
                self.min_impurity_decrease,
                self.store_leaf_values,
            )
        builder.initialize_node_queue(self.tree_, X, y, sample_weight)
        builder.build(self.tree_, X, y, sample_weight)

        self._prune_tree()
        return self

    def fit(self, X, y, sample_weight=None, check_input=True, classes=None):
        """Build a decision tree classifier from the training set (X, y).

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The training input samples. Internally, it will be converted to
            ``dtype=np.float32`` and if a sparse matrix is provided
            to a sparse ``csc_matrix``.

        y : array-like of shape (n_samples,) or (n_samples, n_outputs)
            The target values (class labels) as integers or strings.

        sample_weight : array-like of shape (n_samples,), default=None
            Sample weights. If None, then samples are equally weighted. Splits
            that would create child nodes with net zero or negative weight are
            ignored while searching for a split in each node. Splits are also
            ignored if they would result in any single class carrying a
            negative weight in either child node.

        check_input : bool, default=True
            Allow to bypass several input checking.
            Don't use this parameter unless you know what you're doing.

        classes : array-like of shape (n_classes,), default=None
            List of all the classes that can possibly appear in the y vector.

        Returns
        -------
        self : MultiViewDecisionTreeClassifier
            Fitted estimator.
        """
        return self._fit(
            X, y, sample_weight=sample_weight, check_input=check_input, classes=classes
        )

    def _fit(
        self,
        X,
        y,
        sample_weight=None,
        check_input=True,
        missing_values_in_feature_mask=None,
        classes=None,
    ):
        # XXX: BaseDecisionTree does a check that requires max_features to not be a list/array-like
        # so we need to temporarily set it to an acceptable value
        # in the meantime, we will reset:
        #  - self.max_features_ to the original value
        #  - self.max_features_arr contains a possible array-like setting of max_features
        self._max_features_arr = self.max_features
        self.max_features = None
        super()._fit(X, y, sample_weight, check_input, missing_values_in_feature_mask, classes)
        self.max_features = self._max_features_arr
        return self

    @property
    def _inheritable_fitted_attribute(self):
        """Define additional attributes to pass onto a parent meta tree-estimator.

        Used for passing parameters to HonestTreeClassifier.
        """
        return [
            "max_features_",
            "feature_combinations_",
            "feature_set_ends_",
            "n_feature_sets_",
            "n_features_in_set_",
            "max_features_per_set_",
        ]



"""Scikit manifold oblique random forests."""

import logging
import os
import sys

__version__ = "0.10.2"
logger = logging.getLogger(__name__)


# On OSX, we can get a runtime error due to multiple OpenMP libraries loaded
# simultaneously. This can happen for instance when calling BLAS inside a
# prange. Setting the following environment variable allows multiple OpenMP
# libraries to be loaded. It should not degrade performances since we manually
# take care of potential over-subcription performance issues, in sections of
# the code where nested OpenMP loops can happen, by dynamically reconfiguring
# the inner OpenMP runtime to temporarily disable it while under the scope of
# the outer OpenMP parallel section.
os.environ.setdefault("KMP_DUPLICATE_LIB_OK", "True")

# Workaround issue discovered in intel-openmp 2019.5:
# https://github.com/ContinuumIO/anaconda-issues/issues/11294
os.environ.setdefault("KMP_INIT_AT_FORK", "FALSE")


try:
    # This variable is injected in the __builtins__ by the build
    # process. It is used to enable importing subpackages of sklearn when
    # the binaries are not built
    __treeple_SETUP__  # type: ignore
except NameError:
    __treeple_SETUP__ = False

if __treeple_SETUP__:
    sys.stderr.write("Running from treeple source directory.\n")
    sys.stderr.write("Partial import of treeple during the build process.\n")
    # We are not importing the rest of treeple during the build
    # process, as it may not be compiled yet
else:
    try:
        from . import _lib, tree, ensemble, experimental, stats
        from ._lib.sklearn.ensemble._forest import (
            RandomForestClassifier,
            RandomForestRegressor,
            ExtraTreesClassifier,
            ExtraTreesRegressor,
        )
        from .neighbors import NearestNeighborsMetaEstimator
        from .ensemble import ExtendedIsolationForest, MultiViewRandomForestClassifier
        from .ensemble._unsupervised_forest import (
            UnsupervisedRandomForest,
            UnsupervisedObliqueRandomForest,
        )
        from .ensemble._supervised_forest import (
            ExtraObliqueRandomForestClassifier,
            ExtraObliqueRandomForestRegressor,
            ObliqueRandomForestClassifier,
            ObliqueRandomForestRegressor,
            PatchObliqueRandomForestClassifier,
            PatchObliqueRandomForestRegressor,
        )
        from .ensemble._honest_forest import HonestForestClassifier
    except ImportError as e:
        print(e.msg)
        msg = """Error importing treeple: you cannot import treeple while
        being in treeple source directory; please exit the treeple source
        tree first and relaunch your Python interpreter."""
        raise Exception(e)
        # raise ImportError(msg) from e

    __all__ = [
        "_lib",
        "tree",
        "experimental",
        "ensemble",
        "stats",
        "ExtraObliqueRandomForestClassifier",
        "ExtraObliqueRandomForestRegressor",
        "NearestNeighborsMetaEstimator",
        "ObliqueRandomForestClassifier",
        "ObliqueRandomForestRegressor",
        "PatchObliqueRandomForestClassifier",
        "PatchObliqueRandomForestRegressor",
        "UnsupervisedRandomForest",
        "UnsupervisedObliqueRandomForest",
        "HonestForestClassifier",
        "RandomForestClassifier",
        "RandomForestRegressor",
        "ExtraTreesClassifier",
        "ExtraTreesRegressor",
        "ExtendedIsolationForest",
        "MultiViewRandomForestClassifier",
    ]



"""
===============
Calculating CMI
===============
"""

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from scipy.stats import entropy

from treeple.datasets import make_trunk_classification
from treeple.ensemble import HonestForestClassifier
from treeple.stats import build_oob_forest
from treeple.tree import MultiViewDecisionTreeClassifier

sns.set(color_codes=True, style="white", context="talk", font_scale=1.5)
PALETTE = sns.color_palette("Set1")
sns.set_palette(PALETTE[1:5] + PALETTE[6:], n_colors=9)
sns.set_style("white", {"axes.edgecolor": "#dddddd"})
# %%
# CMI
# ---
#
# Conditional mutual information (*CMI*) measures the dependence of *Y* on
# *X* conditioned on *Z*. It can be calculated by the difference between
# the joint MI (``I([X, Z]; Y)``) and the MI on Z (``I(Y; Z)``):
#
# .. math:: I(X; Y | Z) = I([X, Z]; Y) - I(Y; Z)
#
# With a multiview binary class simulation as an example, this tutorial
# will show how to use ``treeple`` to calculate the statistic with
# multiview data.

# %%
# Create a simulation with two gaussians
# --------------------------------------


# create a binary class simulation with two gaussians
# 500 samples for each class, class zero is standard
# gaussian, and class one has a mean at one for Z
Z, y = make_trunk_classification(
    n_samples=1000,
    n_dim=1,
    mu_0=0,
    mu_1=1,
    n_informative=1,
    seed=1,
)

# class one has a mean at two for X
X, y = make_trunk_classification(
    n_samples=1000,
    n_dim=1,
    mu_0=0,
    mu_1=2,
    n_informative=1,
    seed=2,
)

Z_X = np.hstack((Z, X))


Z_X_y = np.hstack((Z_X, y.reshape(-1, 1)))
Z_X_y = pd.DataFrame(Z_X_y, columns=["Z", "X", "y"])
Z_X_y = Z_X_y.replace({"y": 0.0}, "Class Zero")
Z_X_y = Z_X_y.replace({"y": 1.0}, "Class One")

fig, ax = plt.subplots(figsize=(6, 6))
fig.tight_layout()
ax.tick_params(labelsize=15)
sns.scatterplot(data=Z_X_y, x="Z", y="X", hue="y", palette=PALETTE[:2][::-1], alpha=0.2)
sns.kdeplot(data=Z_X_y, x="Z", y="X", hue="y", palette=PALETTE[:2][::-1], alpha=0.6)
ax.set_ylabel("Variable Two", fontsize=15)
ax.set_xlabel("Variable One", fontsize=15)
plt.legend(frameon=False, fontsize=15)


# %%
# Fit the model with X and Z
# --------------------------


# initialize the forest with 100 trees
est = HonestForestClassifier(
    n_estimators=100,
    max_samples=1.6,
    max_features=0.3,
    bootstrap=True,
    stratify=True,
    tree_estimator=MultiViewDecisionTreeClassifier(),
    random_state=1,
)

# fit the model and obtain the tree posteriors
_, observe_proba = build_oob_forest(est, Z_X, y)

# generate forest posteriors for the two classes
observe_proba = np.nanmean(observe_proba, axis=0)


fig, ax = plt.subplots(figsize=(6, 6))
fig.tight_layout()
ax.tick_params(labelsize=15)

# histogram plot the posterior probabilities for class one
ax.hist(observe_proba[:500][:, 1], bins=50, alpha=0.6, color=PALETTE[1], label="negative")
ax.hist(observe_proba[500:][:, 1], bins=50, alpha=0.3, color=PALETTE[0], label="positive")
ax.set_ylabel("# of Samples", fontsize=15)
ax.set_xlabel("Class One Posterior", fontsize=15)
plt.legend(frameon=False, fontsize=15)
plt.show()

# %%
# Fit the model with Z only
# -------------------------


# initialize the forest with 100 trees
est = HonestForestClassifier(
    n_estimators=100,
    max_samples=1.6,
    max_features=0.3,
    bootstrap=True,
    stratify=True,
    random_state=1,
)

# fit the model and obtain the tree posteriors
_, single_proba = build_oob_forest(est, Z, y)

# generate forest posteriors for the two classes
single_proba = np.nanmean(single_proba, axis=0)


fig, ax = plt.subplots(figsize=(6, 6))
fig.tight_layout()
ax.tick_params(labelsize=15)

# histogram plot the posterior probabilities for class one
ax.hist(single_proba[:500][:, 1], bins=50, alpha=0.6, color=PALETTE[1], label="negative")
ax.hist(single_proba[500:][:, 1], bins=50, alpha=0.3, color=PALETTE[0], label="positive")
ax.set_ylabel("# of Samples", fontsize=15)
ax.set_xlabel("Class One Posterior", fontsize=15)
plt.legend(frameon=False, fontsize=15)
plt.show()

# %%
# Calculate the statistic
# -----------------------


def Calculate_MI(y_true, y_pred_proba):
    # calculate the conditional entropy
    H_YX = np.mean(entropy(y_pred_proba, base=np.exp(1), axis=1))

    # empirical count of each class (n_classes)
    _, counts = np.unique(y_true, return_counts=True)
    # calculate the entropy of labels
    H_Y = entropy(counts, base=np.exp(1))
    return H_Y - H_YX


joint_mi = Calculate_MI(y, observe_proba)
mi = Calculate_MI(y, single_proba)

print("CMI =", round(joint_mi - mi, 2))



from .._lib.sklearn.utils._typedefs cimport float32_t, int32_t, intp_t

# This defines c-importable functions for other cython files

# TODO: remove these files when sklearn merges refactor defining these in pxd files
# https://github.com/scikit-learn/scikit-learn/pull/25606
cdef void sort(float32_t* Xf, intp_t* samples, intp_t n) noexcept nogil



import numbers
from copy import copy

import numpy as np
from sklearn.base import BaseEstimator, MetaEstimatorMixin
from sklearn.exceptions import NotFittedError
from sklearn.neighbors import NearestNeighbors
from sklearn.utils.validation import check_is_fitted, validate_data

from treeple.tree import DecisionTreeClassifier
from treeple.tree._neighbors import _compute_distance_matrix, compute_forest_similarity_matrix


class NearestNeighborsMetaEstimator(MetaEstimatorMixin, BaseEstimator):
    """Meta-estimator for nearest neighbors.

    Uses a decision-tree, or forest model to compute distances between samples
    and then uses the sklearn's nearest-neighbors API to compute neighbors.

    Parameters
    ----------
    estimator : BaseDecisionTree, BaseForest
        The estimator to use for computing distances.
    n_neighbors : int, optional
        Number of neighbors to use by default for kneighbors queries, by default 5.
    radius : float, optional
        Range of parameter space to use by default for radius_neighbors queries, by default 1.0.
    algorithm : str, optional
        Algorithm used to compute the nearest-neighbors, by default 'auto'.
        See :class:`sklearn.neighbors.NearestNeighbors` for details.
    n_jobs : int, optional
        The number of parallel jobs to run for neighbors, by default None.
    """

    def __init__(self, estimator=None, n_neighbors=5, radius=1.0, algorithm="auto", n_jobs=None):
        self.estimator = estimator
        self.n_neighbors = n_neighbors
        self.algorithm = algorithm
        self.radius = radius
        self.n_jobs = n_jobs

    def get_estimator(self):
        if self.estimator is not None:
            return DecisionTreeClassifier(random_state=0)
        else:
            return copy(self.estimator)

    def fit(self, X, y=None):
        """Fit the nearest neighbors estimator from the training dataset.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The training input samples. Internally, it will be converted to
            ``dtype=np.float32`` and if a sparse matrix is provided
            to a sparse ``csc_matrix``.

        y : array-like of shape (n_samples,) or (n_samples, n_outputs)
            The target values, by default None.

        Returns
        -------
        self : object
            Fitted estimator.
        """
        X, y = validate_data(self, X, y, accept_sparse="csc")

        self.estimator_ = self.get_estimator()
        try:
            check_is_fitted(self.estimator_)
        except NotFittedError:
            self.estimator_.fit(X, y)

        self._fit(X, self.n_neighbors)
        return self

    def _fit(self, X, n_neighbors):
        self.neigh_est_ = NearestNeighbors(
            n_neighbors=n_neighbors,
            algorithm=self.algorithm,
            metric="precomputed",
            n_jobs=self.n_jobs,
        )

        # compute the distance matrix
        aff_matrix = compute_forest_similarity_matrix(self.estimator_, X)
        dists = _compute_distance_matrix(aff_matrix)

        # fit the nearest-neighbors estimator
        self.neigh_est_.fit(dists)

    def kneighbors(self, X=None, n_neighbors=None, return_distance=True):
        """Find the K-neighbors of a point.

        Returns indices of and distances to the neighbors of each point.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape (n_queries, n_features), \
            or (n_queries, n_indexed) if metric == 'precomputed', default=None
            Not used, present for API consistency by convention.

        n_neighbors : int, default=None
            Number of neighbors required for each sample. The default is the
            value passed to the constructor.

        return_distance : bool, default=True
            Whether or not to return the distances.

        Returns
        -------
        neigh_dist : ndarray of shape (n_queries, n_neighbors)
            Array representing the lengths to points, only present if
            return_distance=True.

        neigh_ind : ndarray of shape (n_queries, n_neighbors)
            Indices of the nearest points in the population matrix.
        """
        check_is_fitted(self)

        if n_neighbors is None:
            n_neighbors = self.n_neighbors
        elif n_neighbors <= 0:
            raise ValueError("Expected n_neighbors > 0. Got %d" % n_neighbors)
        elif not isinstance(n_neighbors, numbers.Integral):
            raise TypeError(
                "n_neighbors does not take %s value, enter integer value" % type(n_neighbors)
            )

        if X is not None:
            self._fit(X, n_neighbors)

        return self.neigh_est_.kneighbors(n_neighbors=n_neighbors, return_distance=return_distance)

    def radius_neighbors(self, X=None, radius=None, return_distance=True, sort_results=False):
        """Find the neighbors within a given radius of a point or points.

        Return the indices and distances of each point from the dataset
        lying in a ball with size ``radius`` around the points of the query
        array. Points lying on the boundary are included in the results.

        The result points are *not* necessarily sorted by distance to their
        query point.

        Parameters
        ----------
        X : {array-like, sparse matrix} of (n_samples, n_features), default=None
            The query point or points.
            If not provided, neighbors of each indexed point are returned.
            In this case, the query point is not considered its own neighbor.

        radius : float, or array-like of shape (n_samples,) default=None
            Limiting distance of neighbors to return. The default is the value
            passed to the constructor. If an array-like of shape (n_samples),
            then will query for each sample point with a different radius.

        return_distance : bool, default=True
            Whether or not to return the distances.

        sort_results : bool, default=False
            If True, the distances and indices will be sorted by increasing
            distances before being returned. If False, the results may not
            be sorted. If `return_distance=False`, setting `sort_results=True`
            will result in an error.

            .. versionadded:: 0.22

        Returns
        -------
        neigh_dist : ndarray of shape (n_samples,) of arrays
            Array representing the distances to each point, only present if
            `return_distance=True`. The distance values are computed according
            to the ``metric`` constructor parameter.

        neigh_ind : ndarray of shape (n_samples,) of arrays
            An array of arrays of indices of the approximate nearest points
            from the population matrix that lie within a ball of size
            ``radius`` around the query points.

        Notes
        -----
        Because the number of neighbors of each point is not necessarily
        equal, the results for multiple query points cannot be fit in a
        standard data array.
        For efficiency, `radius_neighbors` returns arrays of objects, where
        each object is a 1D array of indices or distances.
        """
        check_is_fitted(self)

        if X is not None:
            n_samples = X.shape[0]
        else:
            n_samples = self.neigh_est_.n_samples_fit_

        if isinstance(radius, numbers.Number):
            radius = [radius] * n_samples

        # now construct nearest neighbor indices and distances within radius
        nn_ind_data = np.zeros((n_samples,), dtype=object)
        nn_dist_data = np.zeros((n_samples,), dtype=object)
        for idx in range(n_samples):
            nn = self.neigh_est_.radius_neighbors(
                X=X, radius=radius[idx], return_distance=return_distance, sort_results=sort_results
            )

            if return_distance:
                nn_ind_data[idx] = nn[0][idx]
                nn_dist_data[idx] = nn[1][idx]
            else:
                nn_ind_data[idx] = nn

        if return_distance:
            return nn_dist_data, nn_ind_data
        return nn_ind_data



import numpy as np
import pytest
from numpy.testing import assert_array_equal
from sklearn import datasets

from treeple.stats import PermutationHonestForestClassifier

# load the iris dataset (n_samples, 4)
# and randomly permute it
iris = datasets.load_iris()
seed = 12345
rng = np.random.default_rng(seed)

# remove third class
iris_X = iris.data[iris.target != 2]
iris_y = iris.target[iris.target != 2]

p = rng.permutation(iris_X.shape[0])
iris_X = iris_X[p]
iris_y = iris_y[p]


def test_permutationforest_errors():
    """Test permutation forest errors when training."""
    n_samples = 10
    est = PermutationHonestForestClassifier(n_estimators=10, random_state=0)

    # covariate index must be an iterable
    with pytest.raises(RuntimeError, match="covariate_index must be an iterable"):
        est.fit(iris_X[:n_samples], iris_y[:n_samples], covariate_index=0)

    # covariate index must be an iterable of ints
    with pytest.raises(RuntimeError, match="Not all covariate_index"):
        est.fit(iris_X[:n_samples], iris_y[:n_samples], covariate_index=[0, 1.0])

    # covariate index must not have numbers greater than
    with pytest.raises(ValueError, match="The length of the covariate index"):
        est.fit(
            iris_X[:n_samples],
            iris_y[:n_samples],
            covariate_index=np.arange(iris_X.shape[1] + 1, dtype=np.intp),
        )


@pytest.mark.parametrize("permute_per_tree", [True, False])
def test_inbag_samples_different_across_forest(permute_per_tree):
    """Test that inbag samples are different across trees."""
    n_estimators = 10
    est = PermutationHonestForestClassifier(
        n_estimators=n_estimators, random_state=0, permute_per_tree=permute_per_tree
    )

    X = iris_X
    y = iris_y
    est.fit(X, y)

    # covariate index when None is all the features
    assert_array_equal(est.covariate_index_, np.arange(X.shape[1], dtype=np.intp))

    # inbag samples should be different across trees when permute_per_tree=True
    permutation_samples_ = est.permutation_indices_
    permutation_samples_ground = permutation_samples_[0]
    assert not all(
        np.array_equal(permutation_samples_ground, permutation_samples_[idx])
        for idx in range(1, n_estimators)
    )



from sklearn.ensemble._iforest import IsolationForest

from treeple.tree import ExtraObliqueDecisionTreeRegressor


class ExtendedIsolationForest(IsolationForest):
    """
    Extended Isolation Forest Algorithm.

    Return the anomaly score of each sample using the ExtendedIsolationForest algorithm

    The ExtendedIsolationForest 'isolates' observations by randomly selecting a
    sparse linear combination of features and then randomly selecting a split
    value between the maximum and minimum values of the selected feature. The
    main difference from the IsolationForest algorithm is how the splits are
    computed.

    Read more in the :ref:`User Guide <isolation_forest>`.

    Parameters
    ----------
    n_estimators : int, default=100
        The number of base estimators in the ensemble.

    max_samples : "auto", int or float, default="auto"
        The number of samples to draw from X to train each base estimator.
            - If int, then draw `max_samples` samples.
            - If float, then draw `max_samples * X.shape[0]` samples.
            - If "auto", then `max_samples=min(256, n_samples)`.

        If max_samples is larger than the number of samples provided,
        all samples will be used for all trees (no sampling).

    contamination : 'auto' or float, default='auto'
        The amount of contamination of the data set, i.e. the proportion
        of outliers in the data set. Used when fitting to define the threshold
        on the scores of the samples.

            - If 'auto', the threshold is determined as in the
              original paper.
            - If float, the contamination should be in the range (0, 0.5].

    max_features : int or float, default=1.0
        The number of features to draw from X to train each base estimator.

            - If int, then draw `max_features` features.
            - If float, then draw `max(1, int(max_features * n_features_in_))` features.

        Note: using a float number less than 1.0 or integer less than number of
        features will enable feature subsampling and leads to a longer runtime.

    bootstrap : bool, default=False
        If True, individual trees are fit on random subsets of the training
        data sampled with replacement. If False, sampling without replacement
        is performed.

    feature_combinations : int, default=2
        The number of features to on average combine to consider at each split.
        For more information, see `ExtraObliqueDecisionTreeRegressor`.

    n_jobs : int, default=None
        The number of jobs to run in parallel for both :meth:`fit` and
        :meth:`predict`. ``None`` means 1 unless in a
        :obj:`joblib.parallel_backend` context. ``-1`` means using all
        processors. See :term:`Glossary <n_jobs>` for more details.

    random_state : int, RandomState instance or None, default=None
        Controls the pseudo-randomness of the selection of the feature
        and split values for each branching step and each tree in the forest.

        Pass an int for reproducible results across multiple function calls.
        See :term:`Glossary <random_state>`.

    verbose : int, default=0
        Controls the verbosity of the tree building process.

    warm_start : bool, default=False
        When set to ``True``, reuse the solution of the previous call to fit
        and add more estimators to the ensemble, otherwise, just fit a whole
        new forest. See :term:`the Glossary <warm_start>`.

        .. versionadded:: 0.21

    Attributes
    ----------
    estimator_ : :class:`~sklearn.tree.ExtraTreeRegressor` instance
        The child estimator template used to create the collection of
        fitted sub-estimators.

        .. versionadded:: 1.2

    estimators_ : list of ExtraTreeRegressor instances
        The collection of fitted sub-estimators.

    estimators_features_ : list of ndarray
        The subset of drawn features for each base estimator.

    estimators_samples_ : list of ndarray
        The subset of drawn samples (i.e., the in-bag samples) for each base
        estimator.

    max_samples_ : int
        The actual number of samples.

    offset_ : float
        Offset used to define the decision function from the raw scores. We
        have the relation: ``decision_function = score_samples - offset_``.
        ``offset_`` is defined as follows. When the contamination parameter is
        set to "auto", the offset is equal to -0.5 as the scores of inliers are
        close to 0 and the scores of outliers are close to -1. When a
        contamination parameter different than "auto" is provided, the offset
        is defined in such a way we obtain the expected number of outliers
        (samples with decision function < 0) in training.

        .. versionadded:: 0.20

    n_features_in_ : int
        Number of features seen during :term:`fit`.

        .. versionadded:: 0.24

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of features seen during :term:`fit`. Defined only when `X`
        has feature names that are all strings.

        .. versionadded:: 1.0

    See Also
    --------
    sklearn.covariance.EllipticEnvelope : An object for detecting outliers in a
        Gaussian distributed dataset.
    sklearn.svm.OneClassSVM : Unsupervised Outlier Detection.
        Estimate the support of a high-dimensional distribution.
        The implementation is based on libsvm.
    sklearn.neighbors.LocalOutlierFactor : Unsupervised Outlier Detection
        using Local Outlier Factor (LOF).

    Notes
    -----
    The implementation is based on an ensemble of ExtraTreeRegressor. The
    maximum depth of each tree is set to ``ceil(log_2(n))`` where
    :math:`n` is the number of samples used to build the tree
    (see (Liu et al., 2008) for more details).

    References
    ----------
    .. [1] Liu, Fei Tony, Ting, Kai Ming and Zhou, Zhi-Hua. "Isolation forest."
           Data Mining, 2008. ICDM'08. Eighth IEEE International Conference on.
    .. [2] Liu, Fei Tony, Ting, Kai Ming and Zhou, Zhi-Hua. "Isolation-based
           anomaly detection." ACM Transactions on Knowledge Discovery from
           Data (TKDD) 6.1 (2012): 3.

    Examples
    --------
    >>> from treeple.ensemble import ExtendedIsolationForest
    >>> X = [[-1.1], [0.3], [0.5], [100]]
    >>> clf = ExtendedIsolationForest(random_state=0).fit(X)
    >>> clf.predict([[0.1], [0], [90]])
    array([ 1,  1, -1])
    """

    def __init__(
        self,
        *,
        n_estimators=100,
        max_samples="auto",
        contamination="auto",
        max_features=1.0,
        bootstrap=False,
        feature_combinations=None,
        n_jobs=None,
        random_state=None,
        verbose=0,
        warm_start=False,
    ):
        super(IsolationForest, self).__init__(
            estimator=ExtraObliqueDecisionTreeRegressor(
                feature_combinations=feature_combinations,
                max_features=1,
                splitter="random",
                random_state=random_state,
            ),
            # here above max_features has no links with self.max_features
            bootstrap=bootstrap,
            bootstrap_features=False,
            n_estimators=n_estimators,
            max_samples=max_samples,
            max_features=max_features,
            warm_start=warm_start,
            n_jobs=n_jobs,
            random_state=random_state,
            verbose=verbose,
        )

        self.feature_combinations = feature_combinations
        self.contamination = contamination



from .._lib.sklearn.tree import (
    DecisionTreeClassifier,
    DecisionTreeRegressor,
    ExtraTreeClassifier,
    ExtraTreeRegressor,
)
from ._classes import (
    ExtraObliqueDecisionTreeClassifier,
    ExtraObliqueDecisionTreeRegressor,
    ObliqueDecisionTreeClassifier,
    ObliqueDecisionTreeRegressor,
    PatchObliqueDecisionTreeClassifier,
    PatchObliqueDecisionTreeRegressor,
    UnsupervisedDecisionTree,
    UnsupervisedObliqueDecisionTree,
)
from ._honest_tree import HonestTreeClassifier
from ._multiview import MultiViewDecisionTreeClassifier
from ._neighbors import compute_forest_similarity_matrix

__all__ = [
    "ExtraObliqueDecisionTreeClassifier",
    "ExtraObliqueDecisionTreeRegressor",
    "compute_forest_similarity_matrix",
    "UnsupervisedDecisionTree",
    "UnsupervisedObliqueDecisionTree",
    "ObliqueDecisionTreeClassifier",
    "ObliqueDecisionTreeRegressor",
    "PatchObliqueDecisionTreeClassifier",
    "PatchObliqueDecisionTreeRegressor",
    "HonestTreeClassifier",
    "DecisionTreeClassifier",
    "DecisionTreeRegressor",
    "ExtraTreeClassifier",
    "ExtraTreeRegressor",
    "MultiViewDecisionTreeClassifier",
]



import numpy as np
import pytest
from numpy.testing import assert_array_equal

from treeple.datasets import (
    approximate_clf_mutual_information,
    approximate_clf_mutual_information_with_monte_carlo,
    make_marron_wand_classification,
    make_quadratic_classification,
    make_trunk_classification,
    make_trunk_mixture_classification,
)
from treeple.datasets.hyppo import MARRON_WAND_SIMS


def test_make_quadratic_classification_v():
    n_samples = 100
    n_features = 5
    x, v = make_quadratic_classification(n_samples, n_features)
    assert all(val in [0, 1] for val in v)
    assert x.shape == (n_samples * 2, n_features)
    assert len(x) == len(v)


@pytest.mark.parametrize(
    "trunk_gen", [make_trunk_classification, make_trunk_mixture_classification]
)
def test_make_trunk_classification_custom_parameters(trunk_gen):
    # Test with custom parameters
    X, y = trunk_gen(
        n_samples=50,
        n_dim=5,
        n_informative=2,
        mu_0=0,
        rho=0.5,
        band_type="ma",
        return_params=False,
    )
    assert X.shape == (50, 5)
    assert y.shape == (50,)


@pytest.mark.parametrize(
    "trunk_gen", [make_trunk_classification, make_trunk_mixture_classification]
)
def test_make_trunk_classification_autoregressive_cov(trunk_gen):
    # Test with default parameters
    n_dim = 10
    n_informative = 10
    rho = 0.5
    data = trunk_gen(
        n_samples=100,
        n_dim=n_dim,
        n_informative=n_informative,
        rho=rho,
        band_type="ar",
        return_params=True,
    )
    cov_list = data[3]
    if trunk_gen == make_trunk_classification:
        assert len(data) == 4
        assert len(data[2]) == 2
        assert len(data[3]) == 2
        assert_array_equal(cov_list[0][0, :], [rho**idx for idx in range(n_dim)])
    elif trunk_gen == make_trunk_mixture_classification:
        assert len(data) == 5
        assert_array_equal(cov_list[0][0, :], [rho**idx for idx in range(n_dim)])
    assert_array_equal(cov_list[0], cov_list[1])
    assert cov_list[0].shape == (n_dim, n_dim)


def test_make_trunk_classification_return_params():
    # Test with return_params=True and uneven number of samples
    n_informative = 5
    X, y, means, covs = make_trunk_classification(
        n_samples=75, n_dim=10, n_informative=n_informative, return_params=True
    )
    assert X.shape == (74, 10), X.shape
    assert y.shape == (74,)
    assert len(means) == 2
    assert len(covs) == 2


@pytest.mark.parametrize(
    "trunk_gen", [make_trunk_classification, make_trunk_mixture_classification]
)
def test_make_trunk_generator_errors(trunk_gen):
    # Test with an invalid band type
    with pytest.raises(ValueError, match=r"Band type .* must be one of"):
        trunk_gen(n_samples=50, rho=0.5, band_type="invalid_band_type")

    # Test with an invalid band type
    with pytest.warns(RuntimeWarning, match="Number of informative dimensions"):
        trunk_gen(n_samples=50, n_dim=10, n_informative=11, rho=0.5)


def test_make_trunk_mixture_errors():
    # Test with an invalid band type
    with pytest.raises(ValueError, match="Mix must be between 0 and 1."):
        make_trunk_mixture_classification(n_samples=50, rho=0.5, mix=2)


def test_make_marron_wand_errors():
    # Test with an invalid band type
    with pytest.raises(ValueError, match="Simulation must be"):
        make_marron_wand_classification(n_samples=50, rho=0.5, simulation=None)


@pytest.mark.parametrize("simulation", [*MARRON_WAND_SIMS.keys()])
def test_make_marron_wand_simulations(simulation):
    # Test with default parameters
    n_samples = 100
    n_dim = 10
    n_informative = 10
    X, y = make_marron_wand_classification(
        n_samples=n_samples,
        n_dim=n_dim,
        n_informative=n_informative,
        simulation=simulation,
    )
    assert X.shape == (n_samples, n_dim)
    assert y.shape == (n_samples,)


def test_approximate_clf_mutual_information_numerically_close():
    mean1 = np.array([1, 1])
    cov1 = np.array([[1, 0.5], [0.5, 1]])

    mean2 = np.array([-1, -1])
    cov2 = np.array([[1, -0.5], [-0.5, 1]])

    means = [mean1, mean2]
    covs = [cov1, cov2]

    result_approximate = approximate_clf_mutual_information(means, covs)
    result_monte_carlo = approximate_clf_mutual_information_with_monte_carlo(means, covs)

    assert np.isclose(
        result_approximate[0], result_monte_carlo[0], atol=5e-2
    ), f"{result_approximate[0]}, {result_monte_carlo[0]}"


@pytest.mark.parametrize(
    "method",
    [make_trunk_classification, make_trunk_mixture_classification],
)
def test_consistent_fixed_seed_trunk_and_mix(method):
    seed = 0
    n_samples = 10
    n_dim = 5
    n_informative = 3
    X1, _ = method(n_samples=n_samples, n_dim=n_dim, n_informative=n_informative, seed=seed)
    X2, _ = method(n_samples=n_samples + 2, n_dim=n_dim + 1, n_informative=n_informative, seed=seed)

    # informative is the same
    assert_array_equal(X1[: n_samples // 2, :n_informative], X2[: n_samples // 2, :n_informative])
    assert_array_equal(
        X1[n_samples // 2 :, :n_informative], X2[n_samples // 2 + 1 : -1, :n_informative]
    )

    # noise is the same
    assert_array_equal(X1[:, n_informative:], X2[:n_samples, n_informative:n_dim])


@pytest.mark.parametrize(
    "method",
    [make_marron_wand_classification],
)
@pytest.mark.parametrize(
    "simulation",
    [*MARRON_WAND_SIMS.keys()],
)
def test_consistent_fixed_seed_marron_wand(method, simulation):
    seed = 0
    n_samples = 10
    n_dim = 5
    n_informative = 3
    X1, _ = method(
        n_samples=n_samples,
        n_dim=n_dim,
        n_informative=n_informative,
        seed=seed,
        simulation=simulation,
    )
    X2, _ = method(
        n_samples=n_samples + 2,
        n_dim=n_dim + 1,
        n_informative=n_informative,
        seed=seed,
        simulation=simulation,
    )

    # informative is the same
    assert_array_equal(X1[: n_samples // 2, :n_informative], X2[: n_samples // 2, :n_informative])
    assert_array_equal(
        X1[n_samples // 2 :, :n_informative], X2[n_samples // 2 + 1 : -1, :n_informative]
    )

    # noise is the same
    assert_array_equal(X1[:, n_informative:], X2[:n_samples, n_informative:n_dim])



# cython: language_level=3
# cython: boundscheck=False, wraparound=False, initializedcheck=False, cdivision=True
import numpy as np
from cython.parallel import prange

cimport numpy as cnp

cnp.import_array()

from libc.math cimport isnan
from libcpp.unordered_set cimport unordered_set

from treeple._lib.sklearn.tree._utils cimport RAND_R_MAX, rand_uniform

from ._utils cimport rand_weighted_binary

from numpy import float32 as DTYPE

TREE_LEAF = -1
TREE_UNDEFINED = -2
cdef intp_t _TREE_LEAF = TREE_LEAF
cdef intp_t _TREE_UNDEFINED = TREE_UNDEFINED


cpdef apply_marginal_tree(
    BaseTree tree,
    object X,
    const intp_t[:] marginal_indices,
    intp_t traversal_method,
    uint8_t use_sample_weight,
    object random_state
):
    """Apply a dataset to a marginalized tree.

    Parameters
    ----------
    tree : Tree
        The tree to apply.
    X : ndarray of shape (n_samples, n_features)
        The dataset to apply.
    marginal_indices : ndarray of shape (n_marginals,)
        The indices of the features to marginalize, which
        are columns in ``X``.
    traversal_method : intp_t
        The traversal method to use. 0 for 'random', 1 for
        'weighted'.
    use_sample_weight : uint8_t
        Whether or not to use the weighted number of samples
        in each node.
    random_state : object
        The random number state.

    Returns
    -------
    out : ndarray of shape (n_samples,)
        The indices of the leaf that each sample falls into.
    """
    # Check input
    if not isinstance(X, np.ndarray):
        raise ValueError("X should be in np.ndarray format, got %s" % type(X))

    if X.dtype != DTYPE:
        raise ValueError("X.dtype should be np.float32, got %s" % X.dtype)

    cdef intp_t n_marginals = marginal_indices.shape[0]

    # sklearn_rand_r random number state
    cdef uint32_t rand_r_state = random_state.randint(0, RAND_R_MAX)

    # define a set of all marginal indices
    cdef unordered_set[intp_t] marginal_indices_map

    # check all marginal indices are valid, and also convert to an unordered map
    for i in range(n_marginals):
        if marginal_indices[i] >= X.shape[1]:
            raise ValueError(
                "marginal_indices must be less than X.shape[1]"
            )

        marginal_indices_map.insert(marginal_indices[i])

    # now we will apply the dataset to the tree
    out = _apply_dense_marginal(
        tree,
        X,
        marginal_indices_map,
        traversal_method,
        use_sample_weight,
        &rand_r_state
    )
    return out


cdef void _resample_split_node(
    BaseTree tree,
    Node* node,
    unordered_set[intp_t] marginal_indices_map,
    const float32_t[:, :] X,
    const float64_t[:, ::1] y,
    const float64_t[:] sample_weight,
) noexcept nogil:
    pass


cdef inline cnp.ndarray _apply_dense_marginal(
    BaseTree tree,
    const float32_t[:, :] X,
    unordered_set[intp_t] marginal_indices_map,
    intp_t traversal_method,
    uint8_t use_sample_weight,
    uint32_t* rand_r_state
):
    """Finds the terminal region (=leaf node) for each sample in X.

    Applies dense dataset to the tree and returns the indices of
    the leaf that each sample falls into. This marginalizes out
    features that are not in the marginal indices.

    Parameters
    ----------
    tree : Tree
        The tree to apply.
    X : const ndarray of shape (n_samples, n_features)
        The data matrix.
    marginal_indices_map : unordered_set[intp_t]
        The indices of the features to marginalize, which
        are columns in ``X``.
    traversal_method : intp_t
        The traversal method to use. 0 for 'random', 1 for
        'weighted'.
    use_sample_weight : uint8_t
        Whether or not to use the weighted number of samples
        in each node.
    rand_r_state : uint32_t
        The random number state.
    """
    # Extract input
    cdef const float32_t[:, :] X_ndarray = X
    cdef intp_t n_samples = X.shape[0]
    cdef float32_t X_i_node_feature

    cdef float32_t n_node_samples, n_right_samples, n_left_samples
    cdef float64_t p_left
    cdef intp_t is_left

    # Initialize output
    cdef intp_t[:] out = np.zeros(n_samples, dtype=np.intp)

    # Initialize auxiliary data-structure
    cdef Node* node = NULL
    cdef intp_t i = 0

    with nogil:
        for i in prange(n_samples):
            node = tree.nodes

            # While node not a leaf
            while node.left_child != _TREE_LEAF:
                # XXX: this will only work for axis-aligned features
                if is_element_present(marginal_indices_map, node.feature):
                    if traversal_method == 1:
                        # if the feature is in the marginal indices, then we
                        # will flip a weighted coin to go down the left, or
                        # right child
                        if use_sample_weight:
                            n_node_samples = node.weighted_n_node_samples
                            n_left_samples = tree.nodes[node.left_child].weighted_n_node_samples
                            n_right_samples = tree.nodes[node.right_child].weighted_n_node_samples
                        else:
                            n_node_samples = node.n_node_samples
                            n_left_samples = tree.nodes[node.left_child].n_node_samples
                            n_right_samples = tree.nodes[node.right_child].n_node_samples

                        # compute the probabilities for going left and right
                        p_left = (<float64_t>n_left_samples / n_node_samples)

                        # randomly sample a direction
                        is_left = rand_weighted_binary(p_left, rand_r_state)

                        if is_left:
                            node = &tree.nodes[node.left_child]
                        else:
                            node = &tree.nodes[node.right_child]
                    else:
                        # traversal method is 0, so it is completely random
                        # and defined by a coin-flip
                        p_left = rand_uniform(0, 1, rand_r_state)
                        if p_left <= 0.5:
                            node = &tree.nodes[node.left_child]
                        else:
                            node = &tree.nodes[node.right_child]
                else:
                    X_i_node_feature = tree._compute_feature(X_ndarray, i, node)
                    # ... and node.right_child != _TREE_LEAF:
                    if isnan(X_i_node_feature):
                        if node.missing_go_to_left:
                            node = &tree.nodes[node.left_child]
                        else:
                            node = &tree.nodes[node.right_child]
                    elif X_i_node_feature <= node.threshold:
                        node = &tree.nodes[node.left_child]
                    else:
                        node = &tree.nodes[node.right_child]

            out[i] = <intp_t>(node - tree.nodes)  # node offset

    return np.asarray(out)


cdef inline intp_t is_element_present(unordered_set[intp_t]& my_set, intp_t element) noexcept nogil:
    """Helper function to check presence of element in set."""
    cdef unordered_set[intp_t].iterator it = my_set.find(element)

    if it != my_set.end():
        return 1
    else:
        return 0



"""
================
Calculating pAUC
================
"""

import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
from sklearn.metrics import roc_auc_score, roc_curve

from treeple.datasets import make_trunk_classification
from treeple.ensemble import HonestForestClassifier
from treeple.stats import build_oob_forest

sns.set(color_codes=True, style="white", context="talk", font_scale=1.5)
PALETTE = sns.color_palette("Set1")
sns.set_palette(PALETTE[1:5] + PALETTE[6:], n_colors=9)
sns.set_style("white", {"axes.edgecolor": "#dddddd"})
# %%
# pAUC@r
# ------
#
# Partial area under the ROC curve (*pAUC*) integrates the true positive
# rates (*TPR*) when the false positive rates (*FPR*) are below a specific
# percentage threshold. Then the value is normalized by that percentage.
#
# .. math:: pAUC@r = \frac{100}{100 - r} \int_{T_r}^\infty \int_{\mathcal{X}} \mathbb{I}\{\eta(X_1) > \eta(X_0) \} dF_1 dF_0
#
# With a binary class simulation as an example, this tutorial will show
# how to use ``treeple`` to calculate the statistic with 90% specificity
# threshold.

# %%
# Create a simulation with two gaussians
# --------------------------------------


# create a binary class simulation with two gaussians
# 500 samples for each class, class zero is standard
# gaussian, and class one has a mean at one
X, y = make_trunk_classification(
    n_samples=1000,
    n_dim=1,
    mu_0=0,
    mu_1=1,
    n_informative=1,
    seed=1,
)


fig, ax = plt.subplots(figsize=(6, 6))
fig.tight_layout()
ax.tick_params(labelsize=15)

# histogram plot the samples
ax.hist(X[:500], bins=50, alpha=0.6, color=PALETTE[1], label="negative")
ax.hist(X[500:], bins=50, alpha=0.3, color=PALETTE[0], label="positive")
ax.set_xlabel("Variable One", fontsize=15)
ax.set_ylabel("Likelihood", fontsize=15)
plt.legend(frameon=False, fontsize=15)
plt.show()


# %%
# Fit the model
# -------------


# initialize the forest with 100 trees
est = HonestForestClassifier(
    n_estimators=100,
    max_samples=1.6,
    max_features=0.3,
    bootstrap=True,
    stratify=True,
    random_state=1,
)

# fit the model and obtain the tree posteriors
_, observe_proba = build_oob_forest(est, X, y)

# generate forest posteriors for the two classes
observe_proba = np.nanmean(observe_proba, axis=0)


fig, ax = plt.subplots(figsize=(6, 6))
fig.tight_layout()
ax.tick_params(labelsize=15)

# histogram plot the posterior probabilities for class one
ax.hist(observe_proba[:500][:, 1], bins=50, alpha=0.6, color=PALETTE[1], label="negative")
ax.hist(observe_proba[500:][:, 1], bins=50, alpha=0.3, color=PALETTE[0], label="positive")
ax.set_ylabel("# of Samples", fontsize=15)
ax.set_xlabel("Class One Posterior", fontsize=15)
plt.legend(frameon=False, fontsize=15)
plt.show()


# %%
# Calculate the statistic
# -----------------------


def Calculate_pAUC(y_true, y_pred_proba, max_fpr=0.1) -> float:
    # check the shape of true labels
    if y_true.squeeze().ndim != 1:
        raise ValueError(f"y_true must be 1d, not {y_true.shape}")

    # find the positive class and calculate fpr and tpr
    if 0 in y_true or -1 in y_true:
        fpr, tpr, thresholds = roc_curve(
            y_true, y_pred_proba[:, 1], pos_label=1, drop_intermediate=False
        )
    else:
        fpr, tpr, thresholds = roc_curve(
            y_true, y_pred_proba[:, 1], pos_label=2, drop_intermediate=False
        )

    fig, ax = plt.subplots(figsize=(6, 6))
    fig.tight_layout()
    ax.tick_params(labelsize=15)
    ax.set_xlim([-0.005, 1.005])
    ax.set_ylim([-0.005, 1.005])
    ax.set_xlabel("False Positive Rate", fontsize=15)
    ax.set_ylabel("True Positive Rate", fontsize=15)

    ax.plot(fpr, tpr, label="ROC curve", color=PALETTE[1])
    # Calculate pAUC at the specific threshold
    pAUC = roc_auc_score(y_true, y_pred_proba[:, 1], max_fpr=max_fpr)

    pos = np.where(fpr == max_fpr)[0][-1]
    ax.fill_between(
        fpr[:pos],
        tpr[:pos],
        color=PALETTE[0],
        alpha=0.6,
        label="pAUC@90 = " + str(round(pAUC, 2)),
        linestyle="--",
    )
    ax.legend(frameon=False, fontsize=15)
    return pAUC


pAUC = Calculate_pAUC(y, observe_proba)
print("pAUC@90 =", round(pAUC, 2))
# sphinx_gallery_thumbnail_number = -1



from ._eiforest import ExtendedIsolationForest
from ._honest_forest import HonestForestClassifier
from ._multiview import MultiViewRandomForestClassifier
from ._supervised_forest import (
    ExtraObliqueRandomForestClassifier,
    ExtraObliqueRandomForestRegressor,
    ObliqueRandomForestClassifier,
    ObliqueRandomForestRegressor,
    PatchObliqueRandomForestClassifier,
    PatchObliqueRandomForestRegressor,
)
from ._unsupervised_forest import UnsupervisedObliqueRandomForest, UnsupervisedRandomForest



from typing import Callable

import numpy as np
from joblib import Parallel, delayed
from numpy.typing import ArrayLike
from sklearn.base import clone
from sklearn.ensemble._base import _partition_estimators
from sklearn.model_selection import StratifiedKFold, train_test_split

from ..tree._classes import DTYPE
from .forest import ForestTestResult, build_oob_forest
from .permuteforest import PermutationHonestForestClassifier
from .utils import METRIC_FUNCTIONS, POSITIVE_METRICS


def _parallel_predict_proba(predict_proba, X, indices_test):
    """
    This is a utility function for joblib's Parallel.

    It can't go locally in ForestClassifier or ForestRegressor, because joblib
    complains that it cannot pickle it when placed there.
    """
    # each tree predicts proba with a list of output (n_samples, n_classes[i])
    prediction = predict_proba(X[indices_test, :], check_input=False)
    return prediction


def build_permutation_forest(
    est,
    perm_est,
    X,
    y,
    covariate_index=None,
    metric="s@98",
    n_repeats=500,
    verbose=False,
    seed=None,
    return_posteriors=True,
    **metric_kwargs,
):
    """Build a hypothesis testing forest using a permutation-forest approach.

    The permutation-forest approach stems from standard permutaiton-testing, where
    each forest is trained on a new permutation of the dataset. The original test
    statistic is computed on the original data. Then the pvalue is computed
    by comparing the original test statistic to the null distribution of the
    test statistic computed from the permuted forests.

    Parameters
    ----------
    est : Forest
        The type of forest to use. Must be enabled with ``bootstrap=True``.
    perm_est : Forest
        The forest to use for the permuted dataset. Should be
        ``PermutationHonestForestClassifier``.
    X : ArrayLike of shape (n_samples, n_features)
        Data.
    y : ArrayLike of shape (n_samples, n_outputs)
        Binary target, so ``n_outputs`` should be at most 1.
    covariate_index : ArrayLike, optional of shape (n_covariates,)
        The index array of covariates to shuffle, by default None.
    metric : str, optional
        The metric to compute, by default "s@98", for sensitivity at
        98% specificity.
    n_repeats : int, optional
        Number of times to bootstrap sample the two forests to construct
        the null distribution, by default 10000. The construction of the
        null forests will be parallelized according to the ``n_jobs``
        argument of the ``est`` forest.
    verbose : bool, optional
        Verbosity, by default False.
    seed : int, optional
        Random seed, by default None.
    return_posteriors : bool, optional
        Whether or not to return the posteriors, by default True.
    **metric_kwargs : dict, optional
        Additional keyword arguments to pass to the metric function.

    Returns
    -------
    observe_stat : float
        The test statistic. To compute the test statistic, take
        ``permute_stat_`` and subtract ``observe_stat_``.
    pvalue : float
        The p-value of the test statistic.
    orig_forest_proba : ArrayLike of shape (n_estimators, n_samples, n_outputs)
        The predicted posterior probabilities for each estimator on their
        out of bag samples.
    perm_forest_proba : ArrayLike of shape (n_estimators, n_samples, n_outputs)
        The predicted posterior probabilities for each of the permuted estimators
        on their out of bag samples.

    References
    ----------
    .. footbibliography::
    """
    rng = np.random.default_rng(seed)
    metric_func: Callable[[ArrayLike, ArrayLike], float] = METRIC_FUNCTIONS[metric]

    if covariate_index is None:
        covariate_index = np.arange(X.shape[1], dtype=int)

    if not isinstance(perm_est, PermutationHonestForestClassifier):
        raise RuntimeError(
            f"Permutation forest must be a PermutationHonestForestClassifier, got {type(perm_est)}"
        )

    # train the original forest on unpermuted data
    est, orig_forest_proba = build_oob_forest(est, X, y, verbose=verbose)
    y_pred_proba_orig = np.nanmean(orig_forest_proba, axis=0)
    observe_test_stat = metric_func(y, y_pred_proba_orig, **metric_kwargs)

    # get the number of jobs
    index_arr = np.arange(X.shape[0], dtype=int).reshape(-1, 1)

    # train many null forests
    X_perm = X.copy()
    null_dist = []
    for _ in range(n_repeats):
        rng.shuffle(index_arr)
        perm_X_cov = X_perm[index_arr, covariate_index]
        X_perm[:, covariate_index] = perm_X_cov

        #
        perm_est = clone(perm_est)
        perm_est.set_params(random_state=rng.integers(0, np.iinfo(np.int32).max))

        perm_est, perm_forest_proba = build_oob_forest(
            perm_est, X_perm, y, verbose=verbose, covariate_index=covariate_index
        )

        y_pred_proba_perm = np.nanmean(perm_forest_proba, axis=0)
        permute_stat = metric_func(y, y_pred_proba_perm, **metric_kwargs)
        null_dist.append(permute_stat)

    # compute pvalue, which note is opposite that of the Coleman approach, since
    # we are testing if the null distribution results in a test statistic greater
    if metric in POSITIVE_METRICS:
        pvalue = (1 + (null_dist >= observe_test_stat).sum()) / (1 + n_repeats)
    else:
        pvalue = (1 + (null_dist <= observe_test_stat).sum()) / (1 + n_repeats)

    forest_result = ForestTestResult(observe_test_stat, permute_stat, None, pvalue, null_dist)
    if return_posteriors:
        return forest_result, orig_forest_proba, perm_forest_proba
    else:
        return forest_result


def build_cv_forest(
    est,
    X,
    y,
    cv=5,
    test_size=0.2,
    verbose=False,
    return_indices=False,
    seed=None,
):
    """Build a hypothesis testing forest using using cross-validation.

    Parameters
    ----------
    est : Forest
        The type of forest to use. Must be enabled with ``bootstrap=True``.
    X : ArrayLike of shape (n_samples, n_features)
        Data.
    y : ArrayLike of shape (n_samples, n_outputs)
        Binary target, so ``n_outputs`` should be at most 1.
    cv : int, optional
        Number of folds to use for cross-validation, by default 5.
    test_size : float, optional
        Proportion of samples per tree to use for the test set, by default 0.2.
    verbose : bool, optional
        Verbosity, by default False.
    return_indices : bool, optional
        Whether or not to return the train and test indices, by default False.
    seed : int, optional
        Random seed, by default None.

    Returns
    -------
    est : Forest
        Fitted forest.
    all_proba_list : list of ArrayLike of shape (n_estimators, n_samples, n_outputs)
        The predicted posterior probabilities for each estimator on their
        out of bag samples. Length of list is equal to the number of splits.
    train_idx_list : list of ArrayLike of shape (n_samples,)
        The training indices for each split.
    test_idx_list : list of ArrayLike of shape (n_samples,)
        The testing indices for each split.
    """
    X = X.astype(np.float32)
    if cv is not None:
        cv = StratifiedKFold(n_splits=cv, shuffle=True, random_state=seed)
        n_splits = cv.get_n_splits()
        train_idx_list, test_idx_list = [], []
        for train_idx, test_idx in cv.split(X, y):
            train_idx_list.append(train_idx)
            test_idx_list.append(test_idx)
    else:
        n_samples_idx = np.arange(len(y))
        train_idx, test_idx = train_test_split(
            n_samples_idx, test_size=test_size, random_state=seed, shuffle=True, stratify=y
        )

        train_idx_list = [train_idx]
        test_idx_list = [test_idx]
        n_splits = 1

    est_list = []
    all_proba_list = []
    for isplit in range(n_splits):
        # clone the estimator to remove all fitted attributes
        est = clone(est)

        X_train, y_train = X[train_idx_list[isplit], :], y[train_idx_list[isplit]]
        # X_test = X[test_idx_list[isplit], :]

        # build forest
        est.fit(X_train, y_train)

        # now evaluate
        X = est._validate_X_predict(X)

        # if we trained a binning tree, then we should re-bin the data
        # XXX: this is inefficient and should be improved to be in line with what
        # the Histogram Gradient Boosting Tree does, where the binning thresholds
        # are passed into the tree itself, thus allowing us to set the node feature
        # value thresholds within the tree itself.
        if est.max_bins is not None:
            X = est._bin_data(X, is_training_data=False).astype(DTYPE)

        # Assign chunk of trees to jobs
        n_jobs, _, _ = _partition_estimators(est.n_estimators, est.n_jobs)

        # avoid storing the output of every estimator by summing them here
        all_proba = Parallel(n_jobs=n_jobs, verbose=verbose)(
            delayed(_parallel_predict_proba)(e.predict_proba, X, test_idx_list[isplit])
            for e in est.estimators_
        )
        posterior_arr = np.full((est.n_estimators, X.shape[0], est.n_classes_), np.nan)
        for itree, (proba) in enumerate(all_proba):
            posterior_arr[itree, test_idx_list[isplit], ...] = proba.reshape(-1, est.n_classes_)

        all_proba_list.append(posterior_arr)
        est_list.append(est)

    if return_indices:
        return est_list, all_proba_list, train_idx_list, test_idx_list

    return est_list, all_proba_list



import numpy as np

cimport numpy as cnp

from .._lib.sklearn.tree._tree cimport BaseTree, Node
from .._lib.sklearn.utils._typedefs cimport float32_t, float64_t, intp_t, uint8_t, uint32_t


cpdef apply_marginal_tree(
    BaseTree tree,
    object X,
    const intp_t[:] marginal_indices,
    intp_t traversal_method,
    uint8_t use_sample_weight,
    object random_state
)



"""
===============================
ExtendedIsolationForest example
===============================

An example using :class:`~treeple.ExtendedIsolationForest` for anomaly
detection, which compares to the :class:`~sklearn.ensemble.IsolationForest` based
on the algorithm in :footcite:`hariri2019extended`.

In the present example we demo two ways to visualize the decision boundary of an
Extended Isolation Forest trained on a toy dataset.

"""

# %%
# Data generation
# ---------------
#
# We generate two clusters (each one containing `n_samples`) by randomly
# sampling the standard normal distribution as returned by
# :func:`numpy.random.randn`. One of them is spherical and the other one is
# slightly deformed.
#
# For consistency with the :class:`~sklearn.ensemble.IsolationForest` notation,
# the inliers (i.e. the gaussian clusters) are assigned a ground truth label `1`
# whereas the outliers (created with :func:`numpy.random.uniform`) are assigned
# the label `-1`.

from copy import copy

import matplotlib.pyplot as plt
import numpy as np
from sklearn.ensemble import IsolationForest
from sklearn.inspection import DecisionBoundaryDisplay
from sklearn.model_selection import train_test_split

from treeple import ExtendedIsolationForest

n_samples, n_outliers = 120, 40
rng = np.random.RandomState(0)
covariance = np.array([[0.5, -0.1], [0.7, 0.4]])
cluster_1 = 0.4 * rng.randn(n_samples, 2) @ covariance + np.array([2, 2])  # general
cluster_2 = 0.3 * rng.randn(n_samples, 2) + np.array([-2, -2])  # spherical
outliers = rng.uniform(low=-4, high=4, size=(n_outliers, 2))

X = np.concatenate([cluster_1, cluster_2, outliers])
y = np.concatenate([np.ones((2 * n_samples), dtype=int), -np.ones((n_outliers), dtype=int)])

X_train, X_test, y_train, y_test = train_test_split(X, y, stratify=y, random_state=42)

# %%
# We can visualize the resulting clusters:

scatter = plt.scatter(X[:, 0], X[:, 1], c=y, s=20, edgecolor="k")
handles, labels = scatter.legend_elements()
plt.axis("square")
plt.legend(handles=handles, labels=["outliers", "inliers"], title="true class")
plt.title("Gaussian inliers with \nuniformly distributed outliers")
plt.show()

# %%
# Training of the model
# ---------------------

extended_clf = ExtendedIsolationForest(max_samples=100, random_state=0, feature_combinations=2)
extended_clf.fit(X_train)

clf = IsolationForest(max_samples=100, random_state=0)
clf.fit(X_train)

# %%
# Plot discrete decision boundary
# -------------------------------
#
# We use the class :class:`~sklearn.inspection.DecisionBoundaryDisplay` to
# visualize a discrete decision boundary. The background color represents
# whether a sample in that given area is predicted to be an outlier
# or not. The scatter plot displays the true labels.

for name, model in zip(["IsoForest", "ExtendedIsoForest"], [clf, extended_clf]):
    disp = DecisionBoundaryDisplay.from_estimator(
        model,
        X,
        response_method="predict",
        alpha=0.5,
    )
    disp.ax_.scatter(X[:, 0], X[:, 1], c=y, s=20, edgecolor="k")
    disp.ax_.set_title(f"Binary decision boundary \nof {name}")
    plt.axis("square")
    plt.legend(handles=handles, labels=["outliers", "inliers"], title="true class")
    plt.show()


# %%
# Plot path length decision boundary
# ----------------------------------
#
# By setting the `response_method="decision_function"`, the background of the
# :class:`~sklearn.inspection.DecisionBoundaryDisplay` represents the measure of
# normality of an observation. Such score is given by the path length averaged
# over a forest of random trees, which itself is given by the depth of the leaf
# (or equivalently the number of splits) required to isolate a given sample.
#
# When a forest of random trees collectively produce short path lengths for
# isolating some particular samples, they are highly likely to be anomalies and
# the measure of normality is close to `0`. Similarly, large paths correspond to
# values close to `1` and are more likely to be inliers.

for name, model in zip(["IsoForest", "ExtendedIsoForest"], [clf, extended_clf]):
    disp = DecisionBoundaryDisplay.from_estimator(
        model,
        X,
        response_method="decision_function",
        alpha=0.5,
    )
    disp.ax_.scatter(X[:, 0], X[:, 1], c=y, s=20, edgecolor="k")
    disp.ax_.set_title(f"Path length decision boundary \nof {name}")
    plt.axis("square")
    plt.legend(handles=handles, labels=["outliers", "inliers"], title="true class")
    plt.colorbar(disp.ax_.collections[1])
    plt.show()

# %%
# Generate Data
# Produce 2-D dataset with a sinusoidal shape and Gaussian noise added on top.

N = 1000
x = np.random.rand(N) * 8 * np.pi
y = np.sin(x) + np.random.randn(N) / 4.0
X = np.array([x, y]).T

fig = plt.figure(figsize=(6, 6))
fig.add_subplot(111)
plt.plot(X[:, 0], X[:, 1], "o", color=[0.5, 0.5, 0.5])

plt.xlim([-5, 30])
plt.ylim([-3.0, 3.0])

plt.show()

# %%

extended_clf = ExtendedIsolationForest(
    max_samples=100, random_state=0, n_estimators=200, feature_combinations=2
)
extended_clf.fit(X)

clf = IsolationForest(max_samples=100, random_state=0, n_estimators=200)
clf.fit(X)

# %%
# Plot discrete decision boundary

for name, model in zip(["IsoForest", "ExtendedIsoForest"], [clf, extended_clf]):
    disp = DecisionBoundaryDisplay.from_estimator(
        model,
        X,
        response_method="decision_function",
        alpha=0.5,
    )
    disp.ax_.scatter(X[:, 0], X[:, 1], c=y, s=15, edgecolor="k")
    disp.ax_.set_title(f"Path length decision boundary \nof {name}")
    plt.colorbar(disp.ax_.collections[1])
    plt.show()

# %%
# Visualize the prediction of each tree within the forest
# -------------------------------------------------------
#
# We can visualize the prediction of each tree within the forest by using a
# circle plot and showing the depth at which each tree isolates a given sample.
# Here, we will evaluate two samples in the sinusoidal plot: one inlier and one
# outlier. The inlier is located at the center of the sinusoidal and the outlier
# is located at the bottom right corner of the plot.

inlier_sample = np.array([10.0, 0.0])
outlier_sample = np.array([-5.0, -3.0])

for name, model in zip(["IsoForest", "ExtendedIsoForest"], [clf, extended_clf]):
    theta = np.linspace(0, 2 * np.pi, len(model.estimators_))
    max_tree_depth = max([model.estimators_[i].get_depth() for i in range(len(model.estimators_))])

    fig = plt.figure()
    ax = plt.subplot(111, projection="polar")

    radii_in = []
    radii_out = []

    # get the depth of each samples
    for radii, sample, color, lw, alpha in zip(
        [radii_in, radii_out], [inlier_sample, outlier_sample], ["b", "r"], [1, 1.3], [1, 0.9]
    ):
        for i in range(len(model.estimators_)):
            # get the max depth of this tree
            max_depth_tree = model.estimators_[i].get_depth()

            leaf_index = model.estimators_[i].apply(sample.reshape(1, -1))
            # get the depth of each tree's leaf node for this sample
            depth = model._decision_path_lengths[i][leaf_index].squeeze()
            radii.append(depth)

        radii = np.array(radii)
        radii = np.sort(radii) / max_tree_depth
        for j in range(len(radii)):
            ax.plot([theta[j], theta[j]], [0, radii[j]], color=color, alpha=alpha, lw=lw)

        if color == "b":
            radii_in = copy(radii)
        else:
            radii_out = copy(radii)

    ax.set(
        title=f"{name}\nNominal: Mean={np.mean(radii_in).round(3)}, "
        f"Var={np.var(radii_in).round(3)}\n"
        f"Anomaly: Mean={np.mean(radii_out).round(3)}, Var={np.var(radii_out).round(3)}",
        xlabel="Anomaly",
    )
    ax.set_xticklabels([])
    ax.axes.get_xaxis().set_visible(False)
    ax.axes.get_yaxis().set_visible(False)
    fig.tight_layout()
    plt.show()

# %%
# References
# ----------
#
# .. footbibliography::



import numpy as np

from treeple.tree import HonestTreeClassifier


def test_honest_tree_pruning():
    """Test honest tree with pruning to ensure no empty leaves."""
    rng = np.random.default_rng(1234)

    n_samples = 1000
    X = rng.standard_normal(size=(n_samples, 100))
    X[n_samples // 2 :] *= -1
    y = [0] * (n_samples // 2) + [1] * (n_samples // 2)

    clf = HonestTreeClassifier(honest_method="prune", max_features="sqrt", random_state=0)
    clf = clf.fit(X, y)

    nonprune_clf = HonestTreeClassifier(
        honest_method="apply", max_features="sqrt", random_state=0, honest_prior="ignore"
    )
    nonprune_clf = nonprune_clf.fit(X, y)

    assert (
        nonprune_clf.tree_.max_depth >= clf.tree_.max_depth
    ), f"{nonprune_clf.tree_.max_depth} <= {clf.tree_.max_depth}"
    # assert np.all(clf.tree_.children_left != -1)

    # Access the original and pruned trees' attributes
    original_tree = nonprune_clf.tree_
    pruned_tree = clf.tree_

    # Ensure the pruned tree has fewer or equal nodes
    assert (
        pruned_tree.node_count < original_tree.node_count
    ), "Pruned tree has more nodes than the original tree"

    # Ensure the pruned tree has no empty leaves
    assert np.all(pruned_tree.value.sum(axis=(1, 2)) > 0), pruned_tree.value.sum(axis=(1, 2))
    # assert np.all(original_tree.value.sum(axis=(1,2)) > 0), original_tree.value.sum(axis=(1,2))
    assert np.all(pruned_tree.value.sum(axis=(1, 2)) > 0) > np.all(
        original_tree.value.sum(axis=(1, 2)) > 0
    )

    # test that the first three nodes are the same, since these are unlikely to be
    # pruned, and should remain invariant.
    #
    # Note: pruning the tree will have the node_ids change since the tree is
    # ordered via DFS.
    for pruned_node_id in range(3):
        pruned_left_child = pruned_tree.children_left[pruned_node_id]
        pruned_right_child = pruned_tree.children_right[pruned_node_id]

        # Check if the pruned node exists in the original tree
        assert (
            pruned_left_child in original_tree.children_left
        ), "Left child node of pruned tree not found in original tree"
        assert (
            pruned_right_child in original_tree.children_right
        ), "Right child node of pruned tree not found in original tree"

        # Check if the node's parameters match for non-leaf nodes
        if pruned_left_child != -1:
            assert (
                pruned_tree.feature[pruned_node_id] == original_tree.feature[pruned_node_id]
            ), "Feature does not match for node {}".format(pruned_node_id)
            assert (
                pruned_tree.threshold[pruned_node_id] == original_tree.threshold[pruned_node_id]
            ), "Threshold does not match for node {}".format(pruned_node_id)
            assert (
                pruned_tree.weighted_n_node_samples[pruned_node_id]
                == original_tree.weighted_n_node_samples[pruned_node_id]
            ), "Weighted n_node samples does not match for node {}".format(pruned_node_id)



# distutils: language=c++
# cython: cdivision=True
# cython: language_level=3
# cython: boundscheck=False
# cython: wraparound=False
# cython: initializedcheck=False

import numpy as np

cimport numpy as cnp

cnp.import_array()

from .._lib.sklearn.tree._utils cimport rand_int, rand_uniform


cdef inline void fisher_yates_shuffle(
    vector_or_memview indices_to_sample,
    intp_t grid_size,
    uint32_t* random_state,
) noexcept nogil:
    """Shuffle the indices in place using the Fisher-Yates algorithm.
    Parameters
    ----------
    indices_to_sample : A C++ vector or 1D memoryview
        The indices to shuffle.
    grid_size : intp_t
        The size of the grid to shuffle. This is explicitly passed in
        to support the templated `vector_or_memview` type, which allows
        for both C++ vectors and Cython memoryviews. Getitng the length
        of both types uses different API.
    random_state : uint32_t*
        The random state.
    """
    cdef intp_t i, j

    # XXX: should this be `i` or `i+1`? for valid Fisher-Yates?
    for i in range(0, grid_size - 1):
        j = rand_int(i, grid_size, random_state)
        indices_to_sample[j], indices_to_sample[i] = \
            indices_to_sample[i], indices_to_sample[j]


cdef inline int rand_weighted_binary(
    float64_t p0,
    uint32_t* random_state
) noexcept nogil:
    """Sample from integers 0 and 1 with different probabilities.

    Parameters
    ----------
    p0 : float64_t
        The probability of sampling 0.
    random_state : uint32_t*
        The random state.
    """
    cdef float64_t random_value = rand_uniform(0.0, 1.0, random_state)

    if random_value < p0:
        return 0
    else:
        return 1

cpdef unravel_index(
    intp_t index,
    cnp.ndarray[intp_t, ndim=1] shape
):
    """Converts a flat index or array of flat indices into a tuple of coordinate arrays.

    Purely used for testing purposes.

    Parameters
    ----------
    index : intp_t
        A flat index.
    shape : numpy.ndarray[intp_t, ndim=1]
        The shape of the array into which the flat indices should be converted.

    Returns
    -------
    numpy.ndarray[intp_t, ndim=1]
        A coordinate array having the same shape as the input `shape`.
    """
    index = np.intp(index)
    shape = np.array(shape)
    coords = np.empty(shape.shape[0], dtype=np.intp)
    cdef const intp_t[:] shape_memview = shape
    cdef intp_t[:] coords_memview = coords
    unravel_index_cython(index, shape_memview, coords_memview)
    return coords


cpdef ravel_multi_index(intp_t[:] coords, const intp_t[:] shape):
    """Converts a tuple of coordinate arrays into a flat index.

    Purely used for testing purposes.

    Parameters
    ----------
    coords : numpy.ndarray[intp_t, ndim=1]
        An array of coordinate arrays to be converted.
    shape : numpy.ndarray[intp_t, ndim=1]
        The shape of the array into which the coordinates should be converted.

    Returns
    -------
    intp_t
        The resulting flat index.

    Raises
    ------
    ValueError
        If the input `coords` have invalid indices.
    """
    return ravel_multi_index_cython(coords, shape)


cdef inline void unravel_index_cython(
    intp_t index,
    const intp_t[:] shape,
    vector_or_memview coords
) noexcept nogil:
    """Converts a flat index into a tuple of coordinate arrays.

    Parameters
    ----------
    index : intp_t
        The flat index to be converted.
    shape : numpy.ndarray[intp_t, ndim=1]
        The shape of the array into which the flat index should be converted.
    coords : intp_t[:] or vector[intp_t]
        A preinitialized array of coordinates to store the result of the
        unraveled `index`.
    """
    cdef intp_t ndim = shape.shape[0]
    cdef intp_t j, size

    for j in range(ndim - 1, -1, -1):
        size = shape[j]
        coords[j] = index % size
        index //= size


cdef inline intp_t ravel_multi_index_cython(
    vector_or_memview coords,
    const intp_t[:] shape
) noexcept nogil:
    """Converts a tuple of coordinate arrays into a flat index in the vectorized dimension.

    Parameters
    ----------
    coords : intp_t[:] or vector[intp_t]
         An array of coordinates to be converted and vectorized into a sinlg
    shape : numpy.ndarray[intp_t, ndim=1]
        The shape of the array into which the coordinates should be converted.

    Returns
    -------
    intp_t
        The resulting flat index.

    Raises
    ------
    ValueError
        If the input `coords` have invalid indices.
    """
    cdef intp_t i, ndim
    cdef intp_t flat_index, index

    ndim = len(shape)

    # Compute flat index
    flat_index = 0
    for i in range(ndim):
        index = coords[i]
        # if index < 0 or index >= shape[i]:
        #     raise ValueError("Invalid index")
        flat_index += index
        if i < ndim - 1:
            flat_index *= shape[i + 1]

    return flat_index



import importlib
import os

import numpy as np
import pytest
import scipy.sparse as sp
from numpy.testing import assert_array_equal

import treeple.stats.utils as utils
from treeple import HonestForestClassifier
from treeple.stats.utils import get_per_tree_oob_samples

seed = 1234
rng = np.random.default_rng(seed)


@pytest.mark.parametrize("bootstrap", [True, False])
def test_get_per_tree_oob_samples(bootstrap):
    n_estimators = 5
    est = HonestForestClassifier(n_estimators=n_estimators, random_state=0, bootstrap=bootstrap)

    X = rng.normal(0, 1, (100, 2))
    X[:50] *= -1
    y = [0, 1] * 50
    samples = np.arange(len(y))
    est.fit(X, y)

    if bootstrap:
        inbag_samples = est.estimators_samples_
        oob_samples = [
            [idx for idx in samples if idx not in inbag_samples[jdx]] for jdx in range(n_estimators)
        ]
        oob_samples_ = get_per_tree_oob_samples(est)
        for itree in range(n_estimators):
            assert len(oob_samples[itree]) > 1
            assert_array_equal(oob_samples_[itree], oob_samples[itree])
    else:
        with pytest.raises(RuntimeError, match="Cannot extract out-of-bag samples"):
            get_per_tree_oob_samples(est)


@pytest.mark.parametrize("use_bottleneck", [True, False])
def test_non_nan_samples(use_bottleneck: bool):
    if use_bottleneck and utils.DISABLE_BN_ENV_VAR in os.environ:
        del os.environ[utils.DISABLE_BN_ENV_VAR]
        importlib.reload(utils)
    else:
        os.environ[utils.DISABLE_BN_ENV_VAR] = "1"
        importlib.reload(utils)

    posterior_array = np.array(
        [
            # tree 1
            [
                [0, 1],
                [np.nan, np.nan],
                [np.nan, np.nan],
            ],
            # tree 2
            [
                [0, 1],
                [np.nan, np.nan],
                [1, 0],
            ],
        ]
    )  # [2, 3, 2]

    expected = np.array([0, 2])
    actual = utils._non_nan_samples(posterior_array)
    np.testing.assert_array_equal(expected, actual)


@pytest.mark.parametrize("use_bottleneck", [True, False])
def test_nanmean_f(use_bottleneck: bool):
    if use_bottleneck and utils.DISABLE_BN_ENV_VAR in os.environ:
        del os.environ[utils.DISABLE_BN_ENV_VAR]
        importlib.reload(utils)
    else:
        os.environ[utils.DISABLE_BN_ENV_VAR] = "1"
        importlib.reload(utils)

    posterior_array = np.array(
        [
            [1, 2, np.nan],
            [3, 4, np.nan],
        ]
    )

    expected = np.array([1.5, 3.5])
    actual = utils.nanmean_f(posterior_array, axis=1)
    np.testing.assert_array_equal(expected, actual)


@pytest.mark.parametrize(
    ("forest_indices", "expected"),
    [
        (np.arange(3), np.array([0.375, 0.75, 0.25])),
        (np.arange(3) + 2, np.array([0.10, 0.05, 0.25])),
        (np.arange(3) + 3, np.array([0.10, 0.45, np.nan])),
    ],
)
def test_get_forest_preds_sparse(
    forest_indices,
    expected,
):

    all_y_pred = sp.csc_matrix(
        np.array(
            [
                [0.50, 0.00, 0.00],
                [0.25, 0.75, 0.00],
                [0.00, 0.00, 0.25],
                [0.10, 0.00, 0.00],
                [0.00, 0.05, 0.00],
                [0.00, 0.85, 0.00],
            ]
        )
    )

    all_y_indicator = sp.csc_matrix(
        np.array(
            [
                [1, 0, 0],
                [1, 1, 0],
                [0, 0, 1],
                [1, 0, 0],
                [0, 1, 0],
                [0, 1, 0],
            ]
        )
    )

    np.testing.assert_array_equal(
        utils._get_forest_preds_sparse(all_y_pred, all_y_indicator, forest_indices),
        expected,
    )
